/* 
 * 'relscan'
 * 
 * scan stations from a list to find busy time
 * 
 * running on RASPBERRY PI 2 and Debian Linux
 *   for use with kenwood TH-F7E
 * 
 * (CAUTION: connecting serial port of raspberry with th-f7 requires
 *   level clipping)
 *
 * */

/* ----------------------------------------------------------------------
    Copyright (C) 2012,2015  Meinolf Braeutigam

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

				Meinolf Braeutigam, Kannenbaeckerstr. 14
				53395 Rheinbach, GERMANY
				meinolf.braeutigam@arcor.de
 * ---------------------------------------------------------------------- */

/*
 * compile with
 * 	'gcc -o relscan relscan.c'
 * or equivalend (no special library needed)
 * 
 * run with
 * 	'relscan'
 * (no arguments on commandline needed, 
 *    but system program 'stty' will be used)
 * 
 * 
 * program action (in short):
 * 
 * - after start connects to th-f7e on serial device 'rigname'
 * - syncs to scan period
 * 
 * - sets th-f7e to first frequency from the list
 * - waits for squelch to stabilize ('s_delay')
 * - reads status of squelch
 * - updates (different) counters for data assembly
 * - repeats with next frequency from list
 * 
 * - at end of list (0 terminated):
 *   - waits for end of scan period ('s_period')
 *   - starts scanning with first frequency from list again
 * 
 * - at end of every hour adds data packet to file 'relscan.dat' with ...
 * 	...headline containing:
 * 		date, time, seconds of the hour, seconds since program start,
 * 		  daylight seconds since program start
 * 	...with a data line for each scanned frequency containing:
 * 		item number, frequency, busy time since start of the hour
 * 		  in seconds, busy time since start of program in seconds,
 * 		  busy time in daylight periods since start of program,
 * 		  the 3 busy times as before but in percent
 * 			
 * program handles signals:
 * 
 * 	SIGUSR1 prints temporary result to 'relscan.tmp'
 * 	SIGUSR2 toggles scan activity (for tests and manual rig setting)
 * 
 * program terminates on any error with some printout (to make it simple)
 * 
 * use <CTRL>C to terminate 
 * 
 * ------------------------------------------------------------------------
 * history/mods (still in 'quck and dirty' status for private use):
 * 
 * mbr20150516	added some more comments
 * 		some code rearrangement 
 * mbr20150116  added item number to output list
 * mbr20150112	version 2015
 * 			modified for raspberry host
 * 			frequency list update
 * mbr20120928	code/data for daylite times added
 * mbr20120924	frequency from list (selected channels only)
 * 		patched date display (month starts with 0 in system time)
 * mbr20120904	first edit (PC version)
 * ------------------------------------------------------------------------
 * todo - list:
 * 
 * - save data (from the beginning of the current hour) at program
 * 	termination
 * 
 * - add frequency dependent squelch setting 
 * ------------------------------------------------------------------------
 * */

/* ************************************************************************
 * the following code has to be edited for individual measurement
 * ************************************************************************
 * */

/* ------------------------------------------------------------------------
 * frequency array, 0 - terminated
 * 	(fix size up to now, max. 99, plus terminator)
 * ------------------------------------------------------------------------
 * */

//  example for the author's qth:

int freq[100] = 
{			// DB0...
   438550000,		// ...GWG Remscheid
   438675000,		// ...MYK Mayen
   438712500,		// ...FSE Euskirchen
   438737500,		// ...KOE Koeln
   438800000,		// ...SJ  Duesseldorf
   438837500,		// ...ERK Erkelenz
   438850000,		// ...DTM Bonn, Telekom
   438862500,		// ...DT  Fuchskaute
   438875000,		// ...GL  Berg. Gladbach
   438975000,		// ...CA  Wuppertal
   439000000,		// ...SK  Koeln
   439025000,		// ...NY  Gummersbach
   439050000,		// ...SG  Koenigswinter
   439100000,		// ...XO  Bergheim
   439150000,		// ...LR  Kahler Asten
   439200000,		// ...MKV Meinerzhagen
   439225000,		// ...RR  Roesrath
   439250000,		// ...UHC Grevenbroich
   439262500,		// ...MHS Oberhausen
   439362500,		// ...EIS Siegen
   439375000,		// ...QP  Juelich
   439400000,		// ...MUH Much
   439425000,		// ...RHB Rheinbach
   0			// (termination)
};

/* the time in seconds the rig needs to scan  a l l  frequencies
 *   from the list (modify when list has changed, assume about
 *   3 .. 4 frequencies per second ) */
int s_period = 6;				/* scan period [sec] */

/* the time in u-seconds for squelch stabilization */
int s_delay = 140000;				/* scan delay [usec] */

/* daylight time limits (in UTC - set host's system time appropriate) 
 *	( >= dt_start ... < dt_stop) */

#define dt_start	6
#define dt_stop		21

/* 'DEBUG 1' makes the program verbose for debugging */
//#define DEBUG 1

/* ************************************************************************
 * the following code should not need modification / editing
 * ************************************************************************
 * */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>

/* rig device */

char rigname[] = "/dev/ttyAMA0";

FILE *rigout;
FILE *rigin;

/* misc buffers */

#define bufl 32
char buf1[bufl];				/* commands */
char buf2[bufl];				/* answers */

int sec_h;					/* seconds in the cur hour */
int sec_tot;					/* seconds since prog start */
int sec_tot_d;					/* seconds in daylight */

int bsy_time[100];				/* busy time in sec */
int bsy_time_sum[100];				/* busy time total ... */ 
int bsy_time_sum_d[100];			/* busy time daylight ... */
						/*   in sec since prog start */
int scan_f = 0;					/* scan inhibit flag */

time_t curtime;

struct tm *cbtimep;				/* current (broken-) time p. */

struct tm obtime;				/* old time data */



/* ------------------------------------------------------------------------
 * set current (local) time in global 'curtime' and 'curbtime'
 * ------------------------------------------------------------------------
 * */
void setcurtime(void)
{
   curtime = time(NULL);
   cbtimep = localtime(&curtime);
}

/* ------------------------------------------------------------------------
 * open rig device
 * 	and set interface params by 'stty'
 * ------------------------------------------------------------------------
 * */
void rig_open(void)
{

   strcpy(buf1,"stty -F ");
   strcat(buf1,rigname);
   strcat(buf1," 9600 raw -echo");
   
#ifdef DEBUG
   printf("set rig comm params:\n\t%s\n",buf1);
#endif
     
   if (system(buf1) == -1)
     {
	fprintf(stderr,"fatal error setting rig comm params\n");
	exit(1);
     }
   
   rigout = fopen(rigname,"w");
   rigin = fopen(rigname,"r");
   if ((rigout == NULL) || (rigin == NULL))
       {
	  perror("rig-file open error");
	  exit(1);
       }
}

/* ------------------------------------------------------------------------
 * send command from buf1 (amended by '\r') to rig
 * ------------------------------------------------------------------------
 * */
void s_command(void)
{
#ifdef DEBUG
   printf(" command to rig: |%s|\n",buf1);
#endif

   fflush(rigin);				/* prepare answer in */

   fflush(rigout);				/* empty output buffer */
   fprintf(rigout,"%s\r",buf1);			/* command out to device */
   fflush(rigout);

}


/* ------------------------------------------------------------------------
 * get answer from rig
 * 	place string in buf2
 * 	return 1 if command in answer
 *             0 otherwise
 * ------------------------------------------------------------------------
 * */
int g_answer(void)
{
   int n;
   
   buf2[0] = '\000';				/* set buffer empty */
   
   n = 0;
   do						/* assemble up to \r or EOF */
     {
	buf2[n] = fgetc(rigin);		
	if ((buf2[n] == '\r') || (buf2[n] == EOF)) break;
     }
   while (n++<(bufl-1));
   
   buf2[n] = '\000';				/* set string terminator */

#ifdef DEBUG
   printf(" got answer: |%s|\n",buf2);
#endif

   if (strstr(buf2,buf1) == NULL) return 0;	/* check for command */
   return 1;					/* command is not part of */

}

/* ------------------------------------------------------------------------
 * set frequency
 * ------------------------------------------------------------------------
 * */
int s_freq(long int Hz)
{
   
   strcpy(buf1,"FQ ");				/* set command head */
   sprintf((char *)&buf1[3],"%011.11ld,5",Hz);	/* add freq info and step */
   
   s_command();
   if (g_answer() == 0)
     {
	fprintf(stderr,"\nfatal error setting frequency\n"
		"\tcommand: %s\n"
		"\tanswer:  %s\n",
		buf1,buf2);
	exit(1);
     }

}

/* ------------------------------------------------------------------------
 * check busy status of band 'band'
 * 	return status 0/1 on success
 * 	return -1 on error (unknown answer)
 * ------------------------------------------------------------------------
 * */
int c_busy(int band)
{
   
   sprintf(buf1,"BY %01.1d",band);
   
   s_command();
   g_answer();

   if (strstr(buf2,",1")) return 1;
   if (strstr(buf2,",0")) return 0;
   
   return -1;
}

/* ------------------------------------------------------------------------
 * sync scan to defined period
 * 
 * ------------------------------------------------------------------------
 * */
void sync_scan(void)
{
   int n;
   
   while(scan_f) usleep(1000000);		/* wait for scan inh. clear */

   do 
     {
	setcurtime();				/* fetch actual time */
	for (n=0; n<60; n=n+s_period)
	  {
	     if (cbtimep->tm_sec == n) return;
	  }
	usleep(100000);
     }
   while(1);

}
   


/* ------------------------------------------------------------------------
 * save busy time with date and time in file
 * 	('relscan.dat' or 'relscan.tmp')
 * ------------------------------------------------------------------------
 * */
void save_bsy_time(int i)
{
   int n;
   FILE *save_f;
   
   if (i == 0)					/* in normal mode */
     {
	if ((save_f=fopen("relscan.dat","a")) == NULL)
	  {
	     perror("fatal error opening accumulating save file");
	     exit(1);
	  }
     }
   else						/* if forced by SIGUSR1 */
     {
	if ((save_f=fopen("relscan.tmp","w")) == NULL)
	  {
	     perror("fatal error opening temporary save file");
	     exit(1);
	  }
     }
   						/* the headline */
   
   fprintf(save_f,"%4.4d %02.2d %02.2d %02.2d %02.2d %4d %6d %6d\n",
	   obtime.tm_year+1900, obtime.tm_mon+1, obtime.tm_mday,
	   obtime.tm_hour, obtime.tm_min,
	   sec_h, sec_tot, sec_tot_d);
   
						/* the data */
   for (n=0; n<100; n++)
     {
	if (freq[n] == 0) break;
	fprintf(save_f,"%2d %8.4f %8d %8d %8d %5.1f %5.1f %5.1f\n",
		n+1,
		(double)freq[n]/1.e6,
		bsy_time[n], bsy_time_sum[n], bsy_time_sum_d[n],
		((double)bsy_time[n]/(double)sec_h)*100.,
		((double)bsy_time_sum[n]/(double)sec_tot)*100.,
		((double)bsy_time_sum_d[n]/(double)sec_tot_d)*100.);
     }
   fprintf(save_f,"\n");			/* separator */
   
   fclose(save_f);
}

/* ------------------------------------------------------------------------
 * print itermediate result
 * 	caused by SIGUSR1
 * ------------------------------------------------------------------------
 * */
void p_data(int i)
{
   save_bsy_time(1);				/* currend data to terminal */
}

/* ------------------------------------------------------------------------
 * toggle scan (start / stop)
 * 	caused by SIGUSR2
 * ------------------------------------------------------------------------
 * */
void toggle_scan(int i)
{
   scan_f = (++scan_f) & 0x1;
}

/* ------------------------------------------------------------------------
 * ------------------------------------------------------------------------
 * */

main(int argc, char *argv[])
{
   int m,n;
   long int Hz;
   time_t time2;

   signal(SIGUSR1, p_data);			/* intermediate result */ 
   signal(SIGUSR2, toggle_scan);		/* toggle scan for tests */ 
   
/* initialize data arrays */
   
   for (n=0; n<100; n++)
     {
	bsy_time[n] = 0;
	bsy_time_sum[n] = 0;
     }

/* prepare rig */
   
   rig_open();
   
   strcpy(buf1,"BC 0");				/* force band 0 */
   s_command();
   if (g_answer() == 0)
     {
	printf("\nfatal error selecting band\n\tgot answer: %s\n",buf2);
	exit(1);
     }

/* prepare scan */
   
   printf("Waiting for sync ...\n");
   
   sync_scan();					/* sync to scan period */
   obtime = *cbtimep;				/* fetch first 'old-'time */

   printf("Scan started ...\n");
   
/* scan in endless loop  - terminated by <ctrl>c (or error) */
   
   do
     {
	
	sec_h += s_period;			/* update seconds counters */
	sec_tot += s_period;
						/* also for daylight times */
	if ((cbtimep->tm_hour >=dt_start ) 
	    && (cbtimep->tm_hour < dt_stop))
	     sec_tot_d += s_period;
	
	time2 = time(NULL);
	for (n=0; n<100; n++)
	  {
	     if (freq[n] == 0) break;		/* list is 0 terminated */
	     s_freq(freq[n]);			/* set frequency */
     	     usleep(s_delay);			/* delay for squelch */
   
	     m = c_busy(0);			/* ... check if busy */
	     if (m < 0)				/* terminate on error */
	       {
		  printf("\nfatal error on busy check\n");
		  exit(1);
	       }
	     bsy_time[n] += s_period*m;		/* ... accumulate busy time */
	     bsy_time_sum[n] += s_period*m;
						/* if in daylight ... */
	     if ((cbtimep->tm_hour >=6 ) && (cbtimep->tm_hour <= 19))
	       bsy_time_sum_d[n] += s_period*m;

#ifdef DEBUG
	     if (m != 0)
	       printf("%8.4f %4d %4d %02.2d:%02.2d:%02.2d\n",
		      (double)freq[n]/1.e6,bsy_time[n],bsy_time_sum[n],
		      cbtimep->tm_hour,cbtimep->tm_min,cbtimep->tm_sec);
#endif	     
	  }
	
	if (cbtimep->tm_hour != obtime.tm_hour) /* ... if in new hour */
	  {
	     save_bsy_time(0);			/* save the data
						     of the passed hour */
	     for (m=0; m<100; m++) bsy_time[m] = 0; /* reset busy_times */
	     sec_h = 0;				/* reset sec counter */
	  }
	
	obtime = *cbtimep;
						/* delay if in same second */
	while(time2 == time(NULL)) usleep(100000);
	
	sync_scan();				/* next period */
     }
   while (1);
   
   exit(0);
}
	


